"""Command line interface for mloq."""
from pathlib import Path

import click

from mloq.api import (
    init_config as _init_config,
    requirements as _requirements,
    setup_project_files,
    setup_push_workflow,
    setup_repository,
)
from mloq.configuration.config_values import MLOQFile
from mloq.configuration.core import set_docker_image


override_opt = click.option(
    "--override/--no-override",
    "-o/ ",
    default=False,
    show_default=True,
    help="If True overwrite existing files. If False ignore files that already exist.",
)
config_file_opt = click.option(
    "--filename",
    "-f",
    "config_file",
    default=None,
    show_default=True,
    help="Name of the repository config file",
    type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
)


output_path_arg = click.argument(
    "output", type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True)
)


@click.group()
def cli():
    pass


@cli.command()
@output_path_arg
@override_opt
@click.option(
    "--filename",
    "-f",
    "filename",
    default="mloq.yml",
    show_default=True,
    help="Name of the generated config file",
)
def init_config(override, filename, output):  # noqa
    """
    Initialize a new config file.

    Parameters:

        OUTPUT: Path where the new mloq.yml template will be written.
    """
    filename = filename if filename != "mloq.yml" else None
    _init_config(output, override, filename)
    click.echo(f"Created new repository config at {Path(output) / filename}")


@cli.command()
@override_opt
@output_path_arg
@config_file_opt
def setup(override, config_file, output):  # noqa
    """
    Set up a new project.

    Parameters:

        OUTPUT: Path where the new mloq.yml template will be written.
    """
    setup_repository(path=output, config_file=config_file, override=override)


@cli.command()
@override_opt
@output_path_arg
@config_file_opt
@click.option(
    "--push",
    "-p",
    type=click.Choice(["python", "dist"], case_sensitive=False),
    help="python creates a workflow for pure python projects | dist creates a workflow "
    "for projects that contain non-python extensions.",
)
def workflows(override, config_file, output, push):  # noqa
    """
    Set up GitHub Actions CI workflow.

    Set up the Github Actions workflow containing the CI pipeline \
    triggered on push and pull request events.

    Parameters:

        OUTPUT: Path where the new mloq.yml template will be written.
    """
    setup_push_workflow(workflow=push, override=override, path=output, config_file=config_file)


@cli.command()
@override_opt
@output_path_arg
@config_file_opt
@click.option("--write", "-w", multiple=True, default=["all"])
@click.option("--opt", "options", multiple=True, default=[None])
@click.option("--install", "-i", multiple=True, default=[None])
def requirements(override, output, config_file, options, write, install):  # noqa
    """
    Write the project requirements files.

    Parameters:

        OUTPUT: Path where the new mloq.yml template will be written.
    """
    if "all" in write:
        dev, lint = True, True
    else:
        dev = "test" in write
        lint = "lint" in write

    options = config_file if options[0] is None else options
    install = install if install[0] is not None else None
    _requirements(
        options=options, override=override, path=output, install=install, test=dev, lint=lint
    )


@cli.command()
@override_opt
@output_path_arg
@config_file_opt
def project_files(config_file, output, override):  # noqa
    """
    Create common project config files.

    project-files creates the following configuration files filled with the target template values:

    \b
    - README.md
    - DCO.md
    - code_of_conduct.md
    - CONTRIBUTING.md
    - pyproject.toml
    - setup.py
    - LICENSE
    - Dockerfile
    - MLproject
    - Makefile

    Parameters:

        OUTPUT: Path of the project where the template files will be written.
    """
    setup_project_files(path=output, template=config_file, override=override)


@MLOQFile.click_command(cli, override_opt, output_path_arg)
def quickstart(output, override, **kwargs):  # noqa
    """
    Interactive initialization of a new project.

    Parameters:

        OUTPUT: Path of the project where the template files will be written.
    """
    MLOQFile.set_target(output)
    if click.confirm("Do you want to generate a mloq.yml file?"):
        MLOQFile.save_config(kwargs, from_kwargs=True)
    config = MLOQFile.to_config(**kwargs)
    config = set_docker_image(config)
    if not override and click.confirm("Do you want to override existing files?"):
        override = True
    setup_repository(path=output, config_file=config, override=override)



from pathlib import Path
import tempfile

import pytest

from mloq.api import setup_repository


@pytest.fixture()
def repo_path():
    return Path(__file__).parent / "test_project"


def test_init_repository(repo_path):
    with tempfile.TemporaryDirectory() as tmp:
        setup_repository(Path(tmp), config_file=Path(__file__).parent / "mloq.yml", override=False)



import click

from mloq.configuration.core import Config, ConfigEntry, ConfigFile, ConfigGroup, MultiChoiceConfig


project_name = ConfigEntry(
    "project_name", prompt_text="Select project name", help="Name of the project",
)

owner = ConfigEntry(
    "owner", prompt_text="Github handle of the project owner", help="Owner of the project",
)

email = ConfigEntry("email", prompt_text="Owner contact email")

author = ConfigEntry(
    "author",
    prompt_text="Author of the project",
    help="Person or entity listed as the project author in setup.py",
    default_prompt=owner,
)

copyright_holder = ConfigEntry(
    "copyright_holder",
    default_prompt=owner,
    help="Owner of the copyright of the project.",
    prompt_text="Copyright holder",
)

project_url = ConfigEntry(
    "project_url",
    default_prompt=(
        lambda: (
            None
            if (owner.value is None and project_name.value is None)
            else f"https://github.com/{owner.value}/{project_name.value}"
        )
    ),
    prompt_text="GitHub project url",
    force_prompt=True,
)

bot_name = ConfigEntry(
    "bot_name",
    prompt_text="Bot's GitHub login to push commits in CI",
    help="Bot's GitHub login to push commits in CI",
    default_prompt=owner,
)

bot_email = ConfigEntry("bot_email", prompt_text="Bot account email", default_prompt=email)
default_branch = ConfigEntry(
    "default_branch",
    force_prompt=True,
    default_prompt="master",
    prompt_text="Default branch of the project",
)
license = ConfigEntry("license", value="MIT", help="Project license type", default_prompt="MIT")
description = ConfigEntry("description", help="Short description of the project")
python_versions = MultiChoiceConfig(
    name="python_versions",
    help="Supported python versions",
    choices=["3.6", "3.7", "3.8", "3.9"],
    default_prompt=["3.6", "3.7", "3.8", "3.9"],
    show_default="3.6, 3.7, 3.8, 3.9",
    force_prompt=True,
)

requirements = MultiChoiceConfig(
    "requirements",
    help="Requirement types of the project as a comma separated list. "
    "Available values: {data-science, data-viz, torch, tensorflow}",
    choices=["data-science", "data-viz", "torch", "tensorflow"],
    default_prompt=["data-science", "data-viz", "pytorch"],
    show_default="data-science, data-viz, pytorch",
    force_prompt=True,
)


workflow = ConfigEntry(
    "workflow",
    help="Push workflow for Github Actions CI",
    default_prompt="python",
    type=click.Choice(["python", "dist", "none"], case_sensitive=False),
    force_prompt=True,
)

TEMPLATE = ConfigGroup(
    "template",
    [
        project_name,
        description,
        owner,
        email,
        author,
        copyright_holder,
        project_url,
        bot_name,
        bot_email,
        license,
        python_versions,
        default_branch,
    ],
)

MLOQFile = ConfigFile(Config([TEMPLATE, requirements, workflow]))

DEFAULT_CONFIG = dict(MLOQFile.to_dict())



"""This module defines the functionality to set up Github Actions workflows programmatically."""
from pathlib import Path
from typing import Union

from mloq.configuration.core import ConfigFile
from mloq.directories import create_github_actions_directories
from mloq.files import push_dist_wkf, push_python_wkf
from mloq.templating import write_template


WORKFLOW_NAMES = {
    "dist": push_dist_wkf,
    "python": push_python_wkf,
}


def setup_workflow_template(workflow, root_path: Path, template, override: bool = False):
    """Add the target workflows to the corresponding .github/workflows repository."""
    create_github_actions_directories(root_path)
    workflows_path = Path(root_path) / ".github" / "workflows"
    # TODO: Check for incompatible workflows
    workflow_file = WORKFLOW_NAMES.get(workflow)
    if workflow.lower() == "none":
        return
    elif workflow_file is None:
        print(f"Workflow {workflow} not defined. Skipping")
    else:
        write_template(workflow_file, params=template, path=workflows_path, override=override)


def setup_push_workflow(
    workflow, config_file: Union[str, Path, dict], path: Union[str, Path], override: bool = False
):
    """Initialize the target workflow."""
    config = config_file if isinstance(config_file, dict) else ConfigFile.read_config(config_file)
    create_github_actions_directories(path)
    wkflow = "none" if workflow is None else workflow
    setup_workflow_template(
        workflow=wkflow, template=config["template"], root_path=path, override=override
    )



from distutils.version import StrictVersion
import os
from pathlib import Path
import sys

import requests


def get_versions(package_name):
    url = "https://test.pypi.org/pypi/%s/json" % (package_name,)
    resp = requests.get(url)
    data = resp.json()
    versions = data["releases"].keys()
    try:
        versions = sorted(versions, key=lambda x: StrictVersion(x))
    except Exception:
        versions = sorted(versions, key=str)
    return versions


def rename_wheels(new_version=None):
    dist_dir = Path(os.getcwd()) / "dist"
    for file in os.listdir(str(dist_dir)):
        version = file.split("-")[1]
        new_file = file.replace(version, new_version)
        os.rename(str(dist_dir / file), str(dist_dir / new_file))


def main():
    project = os.environ.get("PROJECT_NAME", "{{project_name}}")
    last_version = get_versions(project)[-1]
    rename_wheels(last_version)
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""This module defines common functionality for dealing with project requirements."""
import os
from pathlib import Path
from typing import Iterable, Optional, Union

from invoke import run

from mloq.directories import copy_file
from mloq.files import (
    data_science_req,
    data_viz_req,
    File,
    lint_req,
    pytorch_req,
    tensorflow_req,
    test_req,
)


REQUIREMENTS_ALIASES = {
    data_science_req: ["data-science", "datascience", "ds"],
    pytorch_req: ["pytorch", "torch"],
    tensorflow_req: ["tensorflow", "tf"],
    data_viz_req: ["data-visualization", "data-viz", "data-vis", "dataviz", "datavis"],
}


def require_cuda(template: dict) -> bool:
    """Return True if any of the project dependencies require CUDA."""
    if "requirements" not in template:
        return False
    options = template["requirements"]
    tf_alias = REQUIREMENTS_ALIASES[tensorflow_req]
    torch_alias = REQUIREMENTS_ALIASES[pytorch_req]
    for option in options:
        if option in tf_alias or option in torch_alias:
            return True
    return False


def get_aliased_requirements_file(option: str) -> File:
    """Get requirement file from aliased name."""
    for file, valid_alias in REQUIREMENTS_ALIASES.items():
        if option in valid_alias:
            return file
    raise KeyError(f"{option} is not a valid name. Valid aliases are {REQUIREMENTS_ALIASES}")


def read_requirements_file(option: str) -> str:
    """Return the content of the target requirements file form an aliased name."""
    req_file = get_aliased_requirements_file(option)
    with open(req_file.src, "r") as f:
        return f.read()


def compose_requirements(options: Iterable[str]) -> str:
    """
    Return the content requirements.txt file with pinned dependencies.

    The returned string contains the combined dependencies\
     for the different options sorted alphabetically.

    Args:
        options: Iterable containing the aliased names of the target dependencies for the project.

    Returns:
        str containing the pinned versions of all the selected requirements.
    """
    requirements_text = ""
    for i, opt in enumerate(options):
        pref = "\n" if i > 0 else ""  # Ensure one requirement per line
        requirements_text += pref + read_requirements_file(opt)
    # Sort requirements alphabetically
    requirements_text = "\n".join(sorted(requirements_text.split("\n"))).lstrip("\n")
    return requirements_text


def write_project_requirements(
    options, out_path=None, out_name="requirements.txt", override: bool = False
):
    """
    Write the composed requirements.txt file.

    The writen file contains pinned dependencies for the libraries \
    required by the provided options.
    """
    out_path = os.getcwd() if out_path is None else out_path
    file_path = out_path / out_name
    if not override and file_path.exists():
        return
    requirements = compose_requirements(options) if options is not None else ""
    with open(file_path, "w") as f:
        f.write(requirements)


def write_dev_requirements(
    out_path=None, override: bool = False, test: bool = True, lint: bool = True
):
    """Write requirements-lint.txt and requirements-test.txt in the target directory."""
    out_path = Path(os.getcwd() if out_path is None else out_path)
    if lint:
        copy_file(lint_req, out_path, override)
    if test:
        copy_file(test_req, out_path, override)


def write_requirements(
    out_path: Union[str, Path],
    options=None,
    out_name="requirements.txt",
    override: bool = False,
    lint: bool = True,
    test: bool = True,
):
    """Write the different requirements.txt files for the project."""
    out_path = Path(out_path)
    if options is not None:
        write_project_requirements(
            options=options, out_path=out_path, out_name=out_name, override=override
        )
    write_dev_requirements(out_path, override=override, test=test, lint=lint)


def install_requirement_file(path, py3: bool = True):
    """Install the dependencies listed in the target requirements file."""
    python = "python3" if py3 else "python"
    run(f"{python} -m pip install -r {str(path)}")


def install_requirements(
    path: Path,
    requirements: Optional[Union[str, bool, Path]],
    test: Optional[Union[str, bool, Path]],
    lint: Optional[Union[str, bool, Path]],
    py3: bool = True,
):
    """Install the dependencies listed in the target requirements files."""
    if requirements is not None and requirements:
        requirements = "requirements.txt" if isinstance(requirements, bool) else requirements
        requirements = requirements if isinstance(requirements, Path) else path / requirements
        install_requirement_file(requirements, py3=py3)
    if test is not None and test:
        test = "requirements-test.txt" if isinstance(test, bool) else test
        test = test if isinstance(test, Path) else path / test
        install_requirement_file(test, py3=py3)
    if lint is not None and lint:
        lint = "requirements-lint.txt" if isinstance(lint, bool) else lint
        lint = lint if isinstance(lint, Path) else path / lint
        install_requirement_file(lint, py3=py3)


def setup_requirements(
    options,
    path: Path,
    test: bool = True,
    lint: bool = True,
    install_reqs: bool = False,
    install_test: bool = False,
    install_lint: bool = False,
    override: bool = False,
    py3: bool = True,
):
    """Write the requirements file and install them if necessary."""
    write_requirements(out_path=path, options=options, test=test, lint=lint, override=override)
    install_requirements(
        path=path, requirements=install_reqs, test=install_test, lint=install_lint, py3=py3
    )



"""The api module exposes the mloq functionality to the CLI."""
import copy
from pathlib import Path
from typing import Optional, Union

from mloq.configuration.core import ConfigFile
from mloq.directories import copy_file, create_project_directories
from mloq.files import file as new_file, repository, ROOT_PATH_FILES, SCRIPTS, test_main
from mloq.requirements import setup_requirements
from mloq.templating import write_template
from mloq.workflows import setup_push_workflow


def init_config(path: Union[str, Path], override: bool = False, filename=None):
    """Write an empty config file to the target path."""
    repo_file = (
        repository
        if filename is None
        else new_file(repository.src, repository.src.parent, filename)
    )
    copy_file(repo_file, Path(path), override)


def requirements(
    options: Union[str, Path, list, tuple],
    path,
    lint: bool = True,
    test: bool = True,
    install=None,
    override: bool = False,
):
    """Write requirements files and install them if requested."""
    if isinstance(options, (Path, str)):
        options = ConfigFile.read_config(Path(options))["requirements"]
    if isinstance(install, (tuple, list)) and "all" in install or install == "all":
        install_reqs = True
        install_lint = True
        install_test = True
    elif isinstance(install, (tuple, list)):
        install_reqs = "reqs" in install or "requirements" in install
        install_lint = "lint" in install
        install_test = "test" in install
    else:
        install_reqs = False
        install_lint = False
        install_test = False
    setup_requirements(
        options=options,
        path=Path(path),
        test=test,
        lint=lint,
        install_test=install_test,
        install_reqs=install_reqs,
        install_lint=install_lint,
        override=override,
    )


def setup_project_files(path, template: Union[Path, str, dict], override: bool = False):
    """Write the template for common repository config files."""
    if isinstance(template, (Path, str)):
        template = ConfigFile.read_config(Path(template))["template"]
    path = Path(path)
    _template = copy.deepcopy(template)
    _template["project_name"] = _template["project_name"].replace("-", "_")
    project_name = _template["project_name"]
    create_project_directories(project_name=project_name, root_path=path, override=override)
    setup_root_files(template=_template, path=path, override=override)
    tests_path = path / project_name / "tests"
    write_template(test_main, params=_template, path=tests_path, override=override)


def setup_scripts(
    path: Union[str, Path], template: Union[Path, str, dict], override: bool = False
):
    """Initialize CI scripts folder files."""
    path = Path(path)
    path = path if path.name == "scripts" else path / "scripts"
    for file in SCRIPTS:
        write_template(file, params=template, path=path, override=override)


def setup_root_files(
    path: Union[str, Path], template: Union[Path, str, dict], override: bool = False
):
    """Initialize root folder files."""
    for file in ROOT_PATH_FILES:
        write_template(file, params=template, path=path, override=override)


def setup_repository(
    path: Union[str, Path],
    config_file: Optional[Union[Path, str, dict]] = None,
    override: bool = False,
):
    """Initialize the project folder structure and all the filled in boilerplate files."""
    path = Path(path)
    config_file = path / "mloq.yml" if config_file is None else config_file
    config = config_file if isinstance(config_file, dict) else ConfigFile.read_config(config_file)
    template = config["template"]
    setup_project_files(path=path, template=template, override=override)
    setup_push_workflow(config.get("workflow"), path=path, config_file=config, override=override)
    requirements(
        options=config["requirements"], path=path, test=True, lint=True, override=override
    )
    setup_scripts(template=template, path=path, override=override)



import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import click
from ruamel.yaml import load as yaml_load, Loader, YAML as RuamelYAML
from ruamel.yaml.compat import StringIO

from mloq.requirements import require_cuda


TEMPLATE_SCHEMA = {
    "project_name": None,
    "owner": None,
    "author": None,
    "email": None,
    "copyright_holder": None,
}
CONFIG_SCHEMA = {"template": {}}


def get_docker_python_version(params: dict) -> str:
    """Return the highest python version defined for the project."""
    max_version = list(sorted(params["template"]["python_versions"]))[-1]
    version = max_version.replace(".", "")
    return f"py{version}"


def set_docker_image(params) -> dict:
    """
    Add to params the base docker container that will be used to define the project's container.

    If the dependencies require cuda the base image will be gpu friendly.
    """
    if "base_docker_image" in params["template"]:
        return params
    v = get_docker_python_version(params)
    ubuntu_v = "ubuntu20.04" if v in ["py38", "py39"] else "ubuntu18.04"
    image = (
        f"fragiletech/{ubuntu_v}-cuda-11.0-{v}"
        if require_cuda(params)
        else f"fragiletech/{ubuntu_v}-base-{v}"
    )
    params["template"]["base_docker_image"] = image
    return params


def parse_python_versions(params: dict) -> dict:
    """Add Python 3.8 as a default supported version to the config dictionary."""
    if "python_versions" not in params["template"]:
        params["template"]["python_version"] = ["3.8"]
    # TODO (guillemdb): sanitize input and check for valid python version format
    return params


class StringYAML(RuamelYAML):
    def dump(self, data, stream=None, **kw):
        inefficient = False
        if stream is None:
            inefficient = True
            stream = StringIO()
        RuamelYAML.dump(self, data, stream, **kw)
        if inefficient:
            return stream.getvalue()


class BaseEntry:
    def __init__(self, name: str, value=None, default=None):
        self._name = name
        self._value = value
        self._default = default

    def __repr__(self):
        return f"{self.__class__.__name__} {self.name} value: {self.value}"

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return (
            self.main_value()
            if self._value is not None
            else os.environ.get(self.env_var_name, self.default)
        )

    @property
    def default(self):
        if isinstance(self._default, BaseEntry):
            return self._default.value
        elif callable(self._default):
            return self._default()
        else:
            return self._default

    @property
    def env_var_name(self):
        return f"MLOQ_{self.name.upper()}"

    def update(self, value):
        self._value = value

    def is_defined(self):
        return self.value is not None

    def main_value(self):
        if isinstance(self._value, BaseEntry):
            return self._value.value
        elif callable(self._value):
            return self._value()
        else:
            return self._value


class ConfigEntry(BaseEntry):
    def __init__(
        self,
        name: str,
        *name_args,
        value=None,
        prompt_text: str = None,
        hide_input: bool = False,
        show_default: bool = True,
        show_choices: bool = True,
        type=None,
        help: str = None,
        hidden: bool = False,
        prompt_suffix: str = ": ",
        show_envvar: bool = False,
        default_prompt=None,
        force_prompt=False,
        **kwargs,
    ):
        super(ConfigEntry, self).__init__(name=name, value=value, default=default_prompt)
        self._prompt_text, self._help = self._share_text(prompt_text, help, name)
        self._hide_input = hide_input
        self.force_prompt = force_prompt
        self._prompt_suffix = prompt_suffix
        self._type = type
        self._show_default = show_default
        self._show_choices = show_choices
        self._option = self._init_option(
            name_args,
            hide_input=self._hide_input,
            show_default=show_default if default_prompt is None else default_prompt,
            show_choices=show_choices,
            type=self._type,
            help=self._help,
            hidden=hidden,
            show_envvar=show_envvar,
            **kwargs,
        )
        self._prompt = self._init_prompt(
            text=self._prompt_text,
            hide_input=self._hide_input,
            type=self._type,
            prompt_suffix=self._prompt_suffix,
            show_default=self._show_default,
            show_choices=self._show_choices,
        )

    @property
    def option(self):
        return self._option

    def prompt(self):
        return self._prompt()

    def define_value(self, value: None, silent: bool = False):
        if value is not None:
            self._default = value
            self._value = value
            self._prompt = self._init_prompt(
                text=self._prompt_text,
                hide_input=self._hide_input,
                type=self._type,
                prompt_suffix=self._prompt_suffix,
                show_default=self._show_default,
                show_choices=self._show_choices,
            )
        if not silent and (value is None or self.force_prompt):
            value = self.prompt()
        self.update(value)

    def _share_text(self, prompt_text, help, name):
        if prompt_text is None:
            prompt_text = name if help is None else help
        if help is None:
            help = prompt_text
        return prompt_text, help

    def _init_option(
        self,
        name_args,
        type,
        help,
        show_default,
        hide_input,
        hidden,
        show_choices,
        show_envvar,
        **attrs,
    ):
        default = f"${self.env_var_name}" if self.default is None else str(show_default)
        return click.option(
            f"--{self.name.replace('_', '-')}",
            *name_args,
            default=lambda: self.value,
            show_default=default,
            hide_input=hide_input,
            type=type,
            help=help,
            hidden=hidden,
            show_choices=show_choices,
            show_envvar=show_envvar,
            prompt=False,
            confirmation_prompt=False,
            is_flag=None,
            flag_value=None,
            multiple=False,
            count=False,
            allow_from_autoenv=True,
            **attrs,
        )

    def _init_prompt(self, **prompt_kwargs):
        return lambda: click.prompt(default=self.value, **prompt_kwargs)


class MultiChoiceConfig(ConfigEntry):
    def __init__(self, name, choices, **kwargs):
        super(MultiChoiceConfig, self).__init__(name=name, **kwargs)
        self._choices = choices

    @property
    def choices(self):
        return self._choices

    def update(self, value):
        self._value = self._parse_string(value)

    def define_value(self, value: None, silent: bool = False):
        if not silent and (value is None or self.force_prompt):
            value = self.prompt()
        value = self._parse_string(value)
        self.update(value)

    def _parse_string(self, value):
        def filter_str(s):
            return s.lstrip().replace("'", "").replace('"', "").replace("[", "").replace("]", "")

        if isinstance(value, str):
            return [filter_str(s) for s in value.split(",")]
        return value


class ConfigGroup:
    def __init__(
        self,
        name,
        entries: Union[
            List[Union[ConfigEntry, "ConfigGroup"]], Tuple[Union[ConfigEntry, "ConfigGroup"]]
        ],
    ):
        self._entries = {e.name: e for e in entries}
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def entries(self):
        return self._entries

    @property
    def option(self):
        def inner(func):
            for entry in self.entries.values():
                func = entry.option(func)
            return func

        return inner

    @property
    def value(self):
        return self.to_dict()

    def has_entry(self, entry: Union[str, ConfigEntry, "ConfigGroup"]) -> bool:
        def simple_has_entry(e):
            return (entry if isinstance(entry, str) else entry.name) in self._entries

        if isinstance(entry, (ConfigEntry, str)):
            return simple_has_entry(entry)
        elif isinstance(entry, ConfigEntry):
            return any([simple_has_entry(e) for e in entry.entries.values()])

    def delete(self, entry: Union[ConfigEntry, str], key_error: bool = True):
        if not self.has_entry(entry):
            if key_error:
                raise ValueError(f"Only ConfigEntry objects allowed, got {type(entry)} instead.")
            return
        name = entry if isinstance(entry, str) else entry.name
        del self._entries[name]

    def __getitem__(self, item):
        return self._entries[item]

    def __setitem__(self, key, value):
        self._entries[key] = value

    def get(self, item, default=None):
        return self._entries.get(item, default=default)

    def update(self, entry: ConfigEntry):
        if not isinstance(entry, ConfigEntry):
            raise ValueError(f"Only ConfigEntry objects allowed, got {type(entry)} instead.")
        self._entries[entry.name] = entry

    def to_dict(self) -> Dict[str, Any]:
        return {e.name: e.value for e in self._entries.values()}

    def to_kwargs(self, default=None):
        """Return a shallow dictionary containing the names and values of all \
         ConfigEntry objects."""
        kwargs = {} if default is None else default
        for entry in self.entries.values():
            if isinstance(entry, ConfigEntry):
                kwargs[entry.name] = entry.value
            elif isinstance(entry, ConfigGroup):
                for child_ent in entry.entries.values():
                    if isinstance(child_ent, ConfigEntry):
                        kwargs[child_ent.name] = child_ent.value
                    elif isinstance(child_ent, ConfigGroup):
                        kwargs.update(child_ent.to_kwargs(kwargs))
        return kwargs

    def define_value(self, silent: bool = False, **kwargs):
        for name, entry in self._entries.items():
            entry.define_value(kwargs.get(name), silent=silent)


class Config:
    def __init__(
        self,
        entries: Union[
            List[Union[ConfigEntry, ConfigGroup]], Tuple[Union[ConfigEntry, ConfigGroup]]
        ],
    ):
        self._entries = {e.name: e for e in entries}

    def __getitem__(self, item):
        return self._entries[item]

    def __setitem__(self, key, value):
        self._entries[key] = value

    @property
    def entries(self):
        return self._entries

    @property
    def option(self):
        def inner(func):
            for entry in self.entries.values():
                func = entry.option(func)
            return func

        return inner

    def click_command(self, group):
        def wrapped(func):
            @group.command(name=func.__name__)
            @self.option
            def inner(**kwargs):
                self.define_value(**kwargs)
                func(**self.to_kwargs())

            inner.__name__ = func.__name__
            return inner

        return wrapped

    def to_dict(self) -> Dict[str, Any]:
        return {e.name: e.value for e in self._entries.values()}

    def to_kwargs(self, default=None):
        kwargs = {} if default is None else default
        for entry in self.entries.values():
            if isinstance(entry, ConfigEntry):
                kwargs[entry.name] = entry.value
            elif isinstance(entry, ConfigGroup):
                for child_ent in entry.entries.values():
                    if isinstance(child_ent, ConfigEntry):
                        kwargs[child_ent.name] = child_ent.value
                    elif isinstance(child_ent, ConfigGroup):
                        kwargs.update(child_ent.to_kwargs(kwargs))
        return kwargs

    def define_value(self, silent: bool = False, **kwargs):
        """Kwargs is a loaded configuration dict."""
        for name, entry in self._entries.items():
            if isinstance(entry, ConfigGroup):
                entry.define_value(silent=silent, **kwargs.get(entry.name))
            elif isinstance(entry, ConfigEntry):
                entry.define_value(kwargs.get(name), silent=silent)


class BaseWrapper:
    """Generic wrapper to wrap any of the other classes."""

    def __init__(self, data, name: str = "_unwrapped"):
        """
        Initialize a :class:`BaseWrapper`.

        Args:
            data: Object that will be wrapped.
            name: Assign a custom attribute name to the wrapped object.

        """
        setattr(self, name, data)
        self.__name = name

    @property
    def unwrapped(self):
        """Access the wrapped object."""
        return getattr(self, self.__name)

    def __repr__(self):
        return self.unwrapped.__repr__()

    def __call__(self, *args, **kwargs):
        """Call the wrapped class."""
        return self.unwrapped.__call__(*args, **kwargs)

    def __str__(self):
        return self.unwrapped.__str__()

    def __len__(self):
        return self.unwrapped.__len__()

    def __getattr__(self, attr):
        # If a BaseWrapper is being wrapped forward the attribute to it
        try:
            if isinstance(self.unwrapped, BaseWrapper):
                return getattr(self.unwrapped, attr)
            return self.unwrapped.__getattribute__(attr)
        except Exception:
            return self.__getattribute__(attr)


class ConfigFile(BaseWrapper):
    DEFAULT_FILE_NAME = "mloq.yml"

    def __init__(
        self,
        config: Config,
        file: Optional[Union[Path, str]] = None,
        target: Optional[Union[Path, str]] = None,
    ):
        super(ConfigFile, self).__init__(config, name="config")
        self._source = file if file is None else Path(file)
        self._source_is_file = self.validate_path(self.source) and self.source.is_file()
        self._file_name = self.source.name if self._source_is_file else self.DEFAULT_FILE_NAME
        # Load target path data
        self._target = target if target is None else Path(target)
        self._target_is_file = self.validate_path(self.target) and self.target.is_file()
        self._target_name = self.target.name if self._target_is_file else self.DEFAULT_FILE_NAME
        source_is_valid = self.validate_path(self.source)
        self._data = self.read_config(self.source, fail_ok=True) if source_is_valid else None

    def __repr__(self):
        return self.yaml_as_string(self.config.to_dict())

    @staticmethod
    def yaml_as_string(params):
        yaml = StringYAML()
        yaml.indent(sequence=4, offset=2)
        return yaml.dump(params)

    @classmethod
    def read_config(cls, path: Union[Path, str], fail_ok: bool = False) -> dict:
        """Load the project configuration from the target path."""
        from mloq.configuration.config_values import DEFAULT_CONFIG

        if isinstance(path, Path):
            path = path / cls.DEFAULT_FILE_NAME if path.is_dir() else path
        try:
            with open(path, "r") as config:
                params = yaml_load(config.read(), Loader)
        except Exception as e:
            if fail_ok:
                return DEFAULT_CONFIG
            raise e
        return params

    @property
    def source(self):
        if self._source is None:
            return
        elif isinstance(self._source, Path):
            return (
                self._source if self._source.is_file() else self._source / self.DEFAULT_FILE_NAME
            )
        raise ValueError(f"source is not a Path: {self._source}")

    @property
    def target(self):
        if self._target is None:
            return
        elif isinstance(self._target, Path):
            if self._target.is_file():
                return self._target
            elif self._target.is_dir() and isinstance(self.source, Path) and self.source.is_file():
                return self._target / self.source.name
            else:
                return self._target / self.DEFAULT_FILE_NAME

        raise ValueError(f"source is not a Path: {self._target}")

    @property
    def data(self):
        return self._data

    @property
    def file_name(self):
        return self._file_name

    def set_file(self, file: Union[str, Path]):
        self._source = Path(file)

    def set_target(self, target: Union[Path, str]):
        self._target = Path(target)

    def validate_path(self, path: Path):
        path = Path(path) if isinstance(path, str) else path
        if isinstance(path, Path):
            return path.is_file() or (path / self.DEFAULT_FILE_NAME).is_file()
        return False

    @property
    def file_option(self):
        config_file_opt = click.option(
            "--file",
            "-f",
            "file",
            default=None,
            show_default="mloq.yml",
            help="Name of the project config file. Defaults to mloq.yml if filename is a path.",
            type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
        )
        return config_file_opt

    @staticmethod
    def _update_kwargs_from_config(base, new):
        # FIXME: (guillemdb) handle defaults from file when default prompt is specified.
        def recursive_update(orig, updt):
            for k, v in dict(orig).items():
                if k in updt and updt[k] is None:
                    updt[k] = orig[k]
                elif isinstance(v, dict):
                    return recursive_update(v, updt)
            return updt

        return recursive_update(base, new)

    @staticmethod
    def _update_config_from_kwargs(conf, kwargs):
        def recursive_update(src, updt):
            for k, v in dict(src).items():
                if k in updt and updt[k] is not None and not isinstance(v, dict):
                    src[k] = updt[k]
                elif isinstance(v, dict):
                    src[k] = recursive_update(src[k], updt)
            return src

        return recursive_update(conf, kwargs)

    @staticmethod
    def write_yaml(data: dict, path: Union[Path, str]):
        yaml = RuamelYAML()
        yaml.indent(sequence=4, offset=2)
        with open(path, "w") as f:
            yaml.dump(data, f)

    def save_config(self, conf: dict, from_kwargs: bool = False):
        from mloq.configuration.config_values import DEFAULT_CONFIG

        conf = dict() if conf is None else conf
        conf = self._update_config_from_kwargs(DEFAULT_CONFIG, conf) if from_kwargs else conf
        output = self.source if self.target is None else self.target
        self.write_yaml(data=conf, path=output)

    def to_config(self, **kwargs):
        from mloq.configuration.config_values import DEFAULT_CONFIG

        return self._update_config_from_kwargs(DEFAULT_CONFIG, kwargs)

    def parse_config(self, config):
        config = parse_python_versions(config)
        return set_docker_image(config)

    @staticmethod
    def flatten_dict(d):
        new_d = {}
        for k, v in d.items():
            if isinstance(v, dict):
                new_d.update(v)
            else:
                new_d[k] = v
        return new_d

    def click_command(self, group, *options, silent: bool = False):
        def wrapped(func):
            @group.command(name=func.__name__)
            @self.file_option
            @self.option
            def inner(file, *args, **kwargs):
                our_kwargs_keys = self.to_kwargs()
                our_kwargs = {k: v for k, v in kwargs.items() if k in our_kwargs_keys}
                other_kwargs = {k: v for k, v in kwargs.items() if k not in our_kwargs_keys}
                if self.validate_path(file):
                    self.set_file(file)
                config = self.read_config(self.source, fail_ok=True)
                config = self._update_config_from_kwargs(config, our_kwargs)
                self.define_value(**config)
                func(*args, **{**other_kwargs, **self.to_kwargs()})

            for opt in options:
                inner = opt(inner)
            return inner

        return wrapped



"""This module defines common functionality for dealing with directories and files."""
import os
from pathlib import Path
from shutil import copyfile
from typing import Union

from mloq.files import File, init, main, version


def create_empty_file(filepath: Path) -> None:
    """
    Create an empty file in the target path.

    Args:
        filepath: Absolute path to the file that will be created.

    Returns:
        None
    """
    if filepath.exists():
        print(f"file {filepath.name} already exists")
        return
    with open(filepath, "w") as _:
        pass


def read_file(file: File) -> str:
    """Return and string with the content of the provided file."""
    with open(file.src, "r") as f:
        return f.read()


def copy_file(file: File, path: Union[Path, str], override: bool = False) -> None:
    """
    Copy the file from src into dst.

    Args:
        file: File object representing the file that will be copied.
        path: Path to the destination of the copied file.
        override: If False, copy the file if it does not already exists in the \
                  target path. If True, override the target file if it is already present.
    Returns:
        None.
    """
    target = path / file.dst
    if not os.path.isfile(str(target)) or override:
        copyfile(file.src, target)
    else:
        print(f"file {file.name} already exists in {target}")


def create_project_directories(project_name: str, root_path: Path, override: False) -> None:
    """
    Initialize the folder structure of a new Python project.

    It creates a directory for the project containing an empty __init__.py file, and a \
    tests folder inside it with its corresponding __init__.py file. This function will only \
    create new files and folders in case they don't already exist, but it won't override any \
    existing file.

    Args:
        project_name: Name of the project. This is the name of the directory that will be created.
        root_path: Absolute path where the new project folder will be created.
        override: If False, copy the file if it does not already exists in the \
                  target path. If True, override the target file if it is already present.

    Returns:
        None.
    """
    # Project dir
    project_path = root_path / project_name
    os.makedirs(project_path, exist_ok=True)
    copy_file(init, project_path, override)
    copy_file(version, project_path, override)
    copy_file(main, project_path, override)
    # Test dir inside project
    test_path = project_path / "tests"
    os.makedirs(test_path, exist_ok=True)
    copy_file(init, test_path, override)
    # Scripts dir
    scripts_path = root_path / "scripts"
    os.makedirs(scripts_path, exist_ok=True)


def create_github_actions_directories(root_path: Union[str, Path]) -> None:
    """
    Initialize the folder structure for using GitHub actions workflows.

    It creates a .github directory on the target path, with an empty workflows \
    directory inside it, so the .github/workflows directory is ready to add GitHub \
    actions workflows. This function will only create new folders in case they don't already exist.

    Args:
        root_path: Absolute path where the .github folder will be crated.

    Returns:
        None.
    """
    gha_path = Path(root_path) / ".github"
    os.makedirs(gha_path, exist_ok=True)
    workflows_path = gha_path / "workflows"
    os.makedirs(workflows_path, exist_ok=True)



"""This module defines all the different assets accessible from mloq."""
from collections import namedtuple
from pathlib import Path


File = namedtuple("File", "name src dst is_static")


def file(name, path, dst=None, is_static: bool = False):
    """Define a new asset as a File namedtuple."""
    dst = dst if dst is not None else name
    return File(name=name, src=path / name, dst=dst, is_static=is_static)


# Assets paths
ASSETS_PATH = Path(__file__).parent / "assets"
TEMPLATES_PATH = ASSETS_PATH / "templates"
STATIC_FILES_PATH = ASSETS_PATH / "static"
REQUIREMENTS_PATH = ASSETS_PATH / "requirements"
WORKFLOWS_PATH = ASSETS_PATH / "workflows"

# Common files
repository = file("mloq.yml", STATIC_FILES_PATH, is_static=True)
gitignore = file(".gitignore", STATIC_FILES_PATH, is_static=True)
dco = file("DCO.md", STATIC_FILES_PATH, is_static=True)
init = file("__init__.py", STATIC_FILES_PATH, is_static=True)
main = file("__main__.py", STATIC_FILES_PATH, is_static=True)
version = file("version.py", STATIC_FILES_PATH, is_static=True)
dockerfile_aarch64 = file("Dockerfile_aarch64", STATIC_FILES_PATH, is_static=True)
build_manylinux_sh = file("build-manylinux-wheels.sh", STATIC_FILES_PATH, is_static=True)
code_of_conduct = file("code_of_conduct.md", TEMPLATES_PATH)

# Requirements files
data_science_req = file("data-science.txt", REQUIREMENTS_PATH)
data_viz_req = file("data-visualization.txt", REQUIREMENTS_PATH, is_static=True)
pytorch_req = file("pytorch.txt", REQUIREMENTS_PATH, is_static=True)
tensorflow_req = file("tensorflow.txt", REQUIREMENTS_PATH, is_static=True)
lint_req = file("requirements-lint.txt", REQUIREMENTS_PATH, is_static=True)
test_req = file("requirements-test.txt", REQUIREMENTS_PATH, is_static=True)

# Templates
mit_license = file("MIT_LICENSE", TEMPLATES_PATH, "LICENSE")
setup_py = file("setup.py", TEMPLATES_PATH)
test_main = file("test_main.txt", TEMPLATES_PATH, "test_main.py")
rename_wheels = file("rename_testpypi_wheels.py", TEMPLATES_PATH)
dockerfile = file("Dockerfile", TEMPLATES_PATH)
makefile = file("Makefile", TEMPLATES_PATH)
makefile_docker = file("makefile.docker", TEMPLATES_PATH)
mlproject = file("MLproject", TEMPLATES_PATH)
readme = file("README.md", TEMPLATES_PATH)
contributing = file("CONTRIBUTING.md", TEMPLATES_PATH)
pyproject_toml = file("pyproject.toml", TEMPLATES_PATH)

# Workflows
push_python_wkf = file("push-python.yml", WORKFLOWS_PATH, "push.yml")
push_dist_wkf = file("push-dist.yml", WORKFLOWS_PATH, "push.yml")

ROOT_PATH_FILES = [
    dco,
    gitignore,
    code_of_conduct,
    mit_license,
    pyproject_toml,
    setup_py,
    makefile,
    dockerfile,
    mlproject,
    readme,
    contributing,
]

WORKFLOW_FILES = [push_dist_wkf, push_python_wkf]

PYTHON_FILES = [init, main, version]

REQUIREMENTS_FILES = [
    lint_req,
    pytorch_req,
    tensorflow_req,
    data_viz_req,
    data_science_req,
    lint_req,
    test_req,
]

SCRIPTS = [build_manylinux_sh, dockerfile_aarch64, makefile_docker, rename_wheels]

ALL_FILES = ROOT_PATH_FILES + WORKFLOW_FILES + PYTHON_FILES + [repository]

ALL_FILE_PATHS = [str(f.src) for f in ALL_FILES]
